
from .functions import FunctionManager, Function
from .variables import VariableManager
from .comments import Comments
from .data import Data
from .indirect_jumps import IndirectJumps
from .labels import Labels
from .cfg import CFGManager
from .xrefs import XRefManager
from .plugin import KnowledgeBasePlugin
from .sync import SynchronizationManager
from .patches import PatchManager
from .key_definitions import KeyDefinitionManager
from .propagations import Propagations
